#! /usr/bin/python
"""
	Name		: ADF2INF.py
	Author		: David Boddie
	Created		: Wed 18th October 2000
	Updated		: Wed 15th November 2000
	Purpose		: Convert ADFS disc images (ADF) to INF files
	WWW		: http://www.david.boddie.net/Software/Tape2Disc/
"""

def number(size, n):

	# Little endian writing

	s = ""

	while size > 0:
		i = n % 256
		s = s + chr(i)
		n = n / 256
		size = size - 1

	return s


def str2num(size, s):

	i = 0
	n = 0
	while i < size:

		n = n | (ord(s[i]) << (i*8))
		i = i + 1

	return n


def binary(size, n):

	s = size
	i = n
	new = ""
	while (i > 0) & (s > 0):

		if (i & 1)==1:
			new = "1" + new
		else:
			new = "0" + new

		i = i >> 1
		s = s - 1

	if s > 0:
		new = ("0"*s) + new

	return new


def chunk(f, n, data):

	# Chunk ID
	f.write(number(2, n))
	# Chunk length
	f.write(number(4, len(data)))
	# Data
	f.write(data)


def read_tracks(f, inter):

	t = ""

	if inter==0:
		try:
			for i in range(0, ntracks):

				t = t + f.read(nsectors * sector_size)

		except:
			print "Less than %i tracks found." % ntracks
			f.close()
			sys.exit()

	else:

		# Tracks are interleaved (0 80 1 81 2 82 ... 79 159) so rearrange them
		# into the form (0 1 2 3 ... 159)

		try:
			for i in range(0,ntracks):

				if i < (ntracks/2):
					f.seek(i*2*nsectors*sector_size, 0)
					t = t + f.read(nsectors*sector_size)
				else:
					j = i - (ntracks/2)
					f.seek(((j*2)+1)*nsectors*sector_size, 0)
					t = t + f.read(nsectors*sector_size)
		except:
			print "Less than %i tracks found." % ntracks
			f.close()
			sys.exit()

	return t


def read_sectors(adf):

	s = []
	try:

		for i in range(0, ntracks):

			for j in range(0, nsectors):

				s.append(adf.read(sector_size))

	except:

		print "Less than %i tracks x %i sectors found." % (ntracks, nsectors)
		adf.close()
		sys.exit()

	return s


def read_string(offset):

	s = ""
	o = offset
	while ord(sectors[o]) > 32:
		s = s + sectors[o]
		o = o + 1

	return s


def safe(s):

	new = ""
	for i in s:
		if ord(i) <= 32:
			break

		if ord(i) >= 128:
			c = ord(i)^128
			if c > 32:
				new = new + chr(c)
		else:
			new = new + i

	return new


def get_freespace():

	base = 0

	free = []
	p = 0
	while sectors[base+p] != 0:

		free.append(str2num(3, sectors[base+p:base_p+3]))

	name = sectors[sector_size-9:sector_size-4]

	disc_size = str2num(3, sectors[sector_size-4:sector_size-1])

	checksum0 = str2num(1, sectors[sector_size-1])

	base = sector_size

	p = 0
	while sectors[base+p] != 0:

		free.append(str2num(3, sectors[base+p:base_p+3]))

	name = name + sectors[base+sector_size-10:base+sector_size-5]

	disc_id = str2num(2, sectors[base+sector_size-5:base+sector_size-3])

	boot = str2num(1, sectors[base+sector_size-3])

	checksum1 = str2num(1, sectors[base+sector_size-1])


def get_catalogue(disc_type, base):

	global disc_name

	head = base
#	base = sector_size*2
	p = 0

	dir_seq = sectors[head + p]
	dir_start = sectors[head+p+1:head+p+5]
	if (dir_start != "Hugo") & (dir_start != "Nick"):
		print "Not a directory"
		return "", []

	p = p + 5

	files = []

	while ord(sectors[head+p]) != 0:

		old_name = sectors[head+p:head+p+10]
		top_set = 0
		counter = 1
		for i in old_name:
			if (ord(i) & 128) != 0:
				top_set = counter
			counter = counter + 1

		name = safe(sectors[head+p:head+p+10])

		load = str2num(4, sectors[head+p+10:head+p+14])
		exe = str2num(4, sectors[head+p+14:head+p+18])
		length = str2num(4, sectors[head+p+18:head+p+22])

		inddiscadd = sector_size * str2num(3, sectors[head+p+22:head+p+25])

		olddirobseq = str2num(1, sectors[head+p+25])

#		print string.expandtabs("%s\t%s\t%s\t%s" % (name, "("+binary(8, olddirobseq)+")", hex(load), hex(exe)))

		if (load == 0) & (exe == 0) & (top_set > 2):
			lower_dir_name, lower_files = get_catalogue(disc_type, inddiscadd)
			files.append([name, lower_files])
#			if lower_files == []:
#				files.append([name, sectors[inddiscadd:inddiscadd+length], load, exe, length])
#				print name, counter
#			else:
#				files.append([name, lower_files])
#				print name, counter, "is dir"
		else:
			files.append([name, sectors[inddiscadd:inddiscadd+length], load, exe, length])
#			print name, counter

		p = p + 26


	# Go to tail of directory structure (0x200 -- 0x700)

	tail = head + (sector_size*4)

	dir_end = sectors[tail+sector_size-5:tail+sector_size-1]
	if (dir_end != "Hugo") & (dir_end != "Nick"):
		print "Discrepancy in directory structure"
		return "", files

	dir_name = safe(sectors[tail+sector_size-52:tail+sector_size-42])
	parent = sector_size*str2num(3, sectors[tail+sector_size-42:tail+sector_size-39])
	dir_title = safe(sectors[tail+sector_size-39:tail+sector_size-20])

	if parent == head:
		disc_name = dir_title

#	print "Directory title", dir_title
#	print "Directory name ", dir_name

	endseq = sectors[tail+sector_size-6]
	if endseq != dir_seq:
		print "Broken directory,", dir_title
		return dir_name, files

	return dir_name, files


def get_leafname(path):

	pos = string.rfind(path, os.sep)
	if pos != -1:
		return path[pos+1:]
	else:
		return path


def print_catalogue(l, path):

	for i in l:

		name = i[0]
		if type(i[1]) != type([]):
			load, exec_addr, length = i[2], i[3], i[4]
			print string.expandtabs(path+"."+name+"\t"+string.upper(hex(load)[2:])+"\t"+string.upper(hex(exec_addr)[2:])+"\t"+string.upper(hex(length)[2:]), 16)
		else:
			print_catalogue(i[1], path + "." + name)


def extract_files(l, path):

	for i in l:
	
		name = i[0]
		if type(i[1]) != type([]):
			load, exec_addr, length = i[2], i[3], i[4]

			# Create the INF file
			out_file = path + os.sep + name
			inf_file = path + os.sep + name + suffix + "inf"
	
			try:
				out = open(out_file, "wb")
				out.write(i[1])
				out.close()
			except:
				print "Couldn't open the file, %s" % out_file

			try:
				inf = open(inf_file, "w")
				load, exec_addr, length = i[2], i[3], i[4]
				inf.write("$."+name+"\t"+string.upper(hex(load)[2:])+"\t"+string.upper(hex(exec_addr)[2:])+"\t"+string.upper(hex(length)[2:]))
				inf.close()
			except:
				print "Couldn't open the file, %s" % inf_file
		else:
			if name != "$":
				# See if the output directory exists
				try:
					os.listdir(path + os.sep + name)
				except:
					try:
						os.mkdir(path + os.sep + name)
						print "Created directory "+path + os.sep + name
					except:
						print "Directory "+get_leafname(path + os.sep + name)+" already exists."

					extract_files(i[1], path + os.sep + name)
			else:
				extract_files(i[1], path)



import os, string, sys

syntax = "Syntax: ADF2INF.py [-l] <ADF file> <destination path>"
version = "0.14 (Wed 15th November 2000)"

if len(sys.argv) < 3:

	print syntax
	print
	print "ADF2INF version "+version
	print
	print "Take the files stored in the directory given and store them as files with"
	print "corresponding INF files."
	print
	print "If the -l flag is specified then the catalogue of the disc will be printed."
	print
	sys.exit()

# Determine whether the file is to be listed

if sys.argv[1] == "-l":
	if len(sys.argv) < 3:
		print syntax
		print
		sys.exit()
	else:
		listing = 1
		adf_file = sys.argv[2]
else:
	listing = 0
	adf_file = sys.argv[1]
	out_path = sys.argv[2]

if sys.platform == "RISCOS":
	suffix = "/"
else:
	suffix = "."

# Disc properties
if string.lower(adf_file[-4:])==(suffix+"adf"):
	ntracks = 80
	nsectors = 16
	sector_size = 256
	interleave = 0
	disc_type = "adf"
elif string.lower(adf_file[-4:])==(suffix+"adl"):
	ntracks = 160
	nsectors = 16
	sector_size = 256
	interleave = 1
	disc_type = "adl"
else:
	print "Please supply a .adf or .adl file."
	print
	sys.exit()

try:
	adf = open(adf_file, "rb")
except:
	print "Couldn't open the ADF file, %s" % adf_file
	print
	sys.exit()


# Read tracks
sectors = read_tracks(adf, interleave)

# Close the ADF file
adf.close()

disc_name = "Untitled"

# Print catalogue
if listing == 1:
	dir_name, files = get_catalogue(disc_type, 2*sector_size)

	print "Contents of", disc_name,":"
	print

	print_catalogue(files, dir_name)

	# Exit
	sys.exit()


disc_name, files = get_catalogue(disc_type, 2*sector_size)

if disc_name != "$":
	out_path = out_path + os.sep + disc_name

# See if the output directory exists
try:
	os.listdir(out_path)
except:
	try:
		os.mkdir(out_path)
		print "Created directory "+out_path
	except:
		print "Directory "+get_leafname(out_path)+" already exists."
		sys.exit()


extract_files(files, out_path)


# Exit
sys.exit()
